/*********************************************************************************************************
* Software License Agreement (BSD License)                                                               *
* Author: Sebastien Decugis <sdecugis@nict.go.jp>							 *
*													 *
* Copyright (c) 2008, WIDE Project and NICT								 *
* All rights reserved.											 *
* 													 *
* Redistribution and use of this software in source and binary forms, with or without modification, are  *
* permitted provided that the following conditions are met:						 *
* 													 *
* * Redistributions of source code must retain the above 						 *
*   copyright notice, this list of conditions and the 							 *
*   following disclaimer.										 *
*    													 *
* * Redistributions in binary form must reproduce the above 						 *
*   copyright notice, this list of conditions and the 							 *
*   following disclaimer in the documentation and/or other						 *
*   materials provided with the distribution.								 *
* 													 *
* * Neither the name of the WIDE Project or NICT nor the 						 *
*   names of its contributors may be used to endorse or 						 *
*   promote products derived from this software without 						 *
*   specific prior written permission of WIDE Project and 						 *
*   NICT.												 *
* 													 *
* THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED *
* WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A *
* PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR *
* ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT 	 *
* LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS 	 *
* INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR *
* TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF   *
* ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.								 *
*********************************************************************************************************/

/* This file contains the definitions of types and functions related to the log facility
 and that can be called by extensions. */
 
#ifndef _LOG_API_H
#define _LOG_API_H

#include <stddef.h>

/*
 ***************************************************************************
 *
 * Log API functions 
 *
 * These functions are used to report information to the user 
 * (debug, errors, ...)
 *
 ***************************************************************************
 */

#ifndef IN_EXTENSION

/*
 * FUNCTION:	log_error
 *
 * PARAMETERS:
 *  format 	: Same format string as in the printf function
 *  ...		: Same list as printf
 *
 * DESCRIPTION: 
 *  Log an error message to the user. Error conditions must be reported with this function.
 *
 * RETURN VALUE:
 *  None.
 */
void log_error ( char * format, ... );

/*
 * FUNCTION:	log_normal
 *
 * PARAMETERS:
 *  format 	: Same format string as in the printf function
 *  ...		: Same list as printf
 *
 * DESCRIPTION: 
 *  Log a message to the user. Warnings and informations are reported with this function
 *
 * RETURN VALUE:
 *  None.
 */
void log_normal ( char * format, ... );

/*
 * FUNCTION:	log_debug
 *
 * PARAMETERS:
 *  format 	: Same format string as in the printf function
 *  ...		: Same list as printf
 *
 * DESCRIPTION: 
 *  Log internal information for use of developpers only. 
 *
 * RETURN VALUE:
 *  None.
 */
void log_debug ( char * format, ... );

#endif /* ! IN_EXTENSION */

/* The version of this API, to check binary compatibility -- increment each time a change is made in api_log_t */
#define WAAAD_API_LOG_VER	1

/* Now define the type of the structure that contains the callback to pass to extensions */
typedef struct {
	/* The header is common to all sub-API pieces */
	size_t	length;		/* The size of this structure, may be useful for extensions not using the log facility */
	int	version;	/* The version of this API/ABI, must be WAAAD_API_LOG_VER */
	
	/* the remaining is log-specific */
	void (*log_error)  ( char * format, ... );
	void (*log_normal) ( char * format, ... );
	void (*log_debug)  ( char * format, ... );
} api_log_t;

#ifdef IN_EXTENSION

/* From within the extensions, we register callbacks in the following global structure */
#ifdef DECLARE_API_POINTERS
api_log_t * g_api_log=NULL;
#else /* DECLARE_API_POINTERS */
extern api_log_t * g_api_log;
#endif /* DECLARE_API_POINTERS */

/* These defines allow to call functions from extension in the same way as in waaad */
#define log_error	g_api_log->log_error
#define log_normal	g_api_log->log_normal
#define log_debug	g_api_log->log_debug


#else /* IN_EXTENSION */

/* From the daemon, we must initialize the API object, in extension.c */
# define MY_WAAAD_API_LOG_VER 1
# if MY_WAAAD_API_LOG_VER != WAAAD_API_LOG_VER
#  error "You must update API_INIT_LOG also"
# endif

#define API_INIT_LOG( api_log ) 			\
{							\
	(api_log).length     = sizeof(api_log_t);	\
	(api_log).version    = WAAAD_API_LOG_VER;	\
	(api_log).log_error  = log_error;		\
	(api_log).log_normal = log_normal;		\
	(api_log).log_debug  = log_debug;		\
}							

#endif /* IN_EXTENSION */

#endif /* ! _LOG_API_H */
