/*********************************************************************************************************
* Software License Agreement (BSD License)                                                               *
* Author: Sebastien Decugis <sdecugis@nict.go.jp>							 *
*													 *
* Copyright (c) 2008, WIDE Project and NICT								 *
* All rights reserved.											 *
* 													 *
* Redistribution and use of this software in source and binary forms, with or without modification, are  *
* permitted provided that the following conditions are met:						 *
* 													 *
* * Redistributions of source code must retain the above 						 *
*   copyright notice, this list of conditions and the 							 *
*   following disclaimer.										 *
*    													 *
* * Redistributions in binary form must reproduce the above 						 *
*   copyright notice, this list of conditions and the 							 *
*   following disclaimer in the documentation and/or other						 *
*   materials provided with the distribution.								 *
* 													 *
* * Neither the name of the WIDE Project or NICT nor the 						 *
*   names of its contributors may be used to endorse or 						 *
*   promote products derived from this software without 						 *
*   specific prior written permission of WIDE Project and 						 *
*   NICT.												 *
* 													 *
* THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED *
* WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A *
* PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR *
* ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT 	 *
* LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS 	 *
* INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR *
* TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF   *
* ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.								 *
*********************************************************************************************************/

/* Conf facility.
 *
 * This file handles the global configuration of the daemon.
 *
 * See conf.h for more information on the functions 
 *
 */

/* NOTE:
 *
 * In case a "reload configuration" function has to be written, this file may need to be made re-entrant.
 * It will require work on every place in the code of the daemon that uses "g_conf" and "g_pconf".
 */

#include "waaad-internal.h"
#include <string.h>
#include <errno.h>
#include <stdio.h>
#include <arpa/inet.h>
#include <netdb.h>
 
waaad_conf_t * g_conf=NULL;

static waaad_conf_t waaad_config;

/* Set the default values in a configuration structure */
static int conf_setdefaults(waaad_conf_t * conf)
{
	/* Clear the configuration */
	memset(conf, 0, sizeof(waaad_conf_t));
	
	/* log to console */
	conf->log_to = LOG_TO_CONSOLE;
	
	/* The default configuration file path (from configure) */
	conf->filepath = DEFAULT_CONF_FILE;
	
	/* The default port */
	conf->pub.local_port = 3868;
	
	/* The default tctimer and twtimer values */
	conf->pub.tctimer = 30;
	conf->pub.twtimer = 30;
	
	return 0;
}

/* dump a list of SA */
static void conf_dump_addr(char * name, sSS * array)
{
	sSS * ptr = array;
	if (!array) {
		TRACE_DEBUG( INFO, "  %-11s : -", name);
	} else {
		TRACE_DEBUG( INFO, "  %-11s :", name);
		while (ptr->ss_family != 0) {
			switch (ptr->ss_family) {
				case AF_INET:
					{
						char buf[INET_ADDRSTRLEN];
						sSA4 * sin = (sSA4 *)ptr;
						if (inet_ntop(sin->sin_family, &sin->sin_addr, buf, sizeof(buf)) == NULL) {
							TRACE_DEBUG( INFO, "                (error in inet_ntop: %s)", strerror(errno));
						} else {
							TRACE_DEBUG( INFO, "                %s", &buf);
						}
					}
					break;
					
				case AF_INET6:
					{
						char buf[INET6_ADDRSTRLEN];
						sSA6 * sin6 = (sSA6 *)ptr;
						if (inet_ntop(sin6->sin6_family, &sin6->sin6_addr, buf, sizeof(buf)) == NULL) {
							TRACE_DEBUG( INFO, "                (error in inet_ntop: %s)", strerror(errno));
						} else {
							TRACE_DEBUG( INFO, "                %s", &buf);
						}
					}
					break;
					
				default:
					TRACE_DEBUG( INFO, "                (invalid AF: %d)", ptr->ss_family)
			}
			ptr++;
		}
	}
}

/* Output the parsed configuration */
static void conf_dump(void)
{
	char buf4[INET_ADDRSTRLEN];
	char buf6[INET6_ADDRSTRLEN];
	char *str4 = "-";
	char *str6 = "-";
	if (g_pconf->local_addr_diamid[0].ss_family == AF_INET) {
		sSA4 * sin = (sSA4 *)&g_pconf->local_addr_diamid[0];
		if (inet_ntop(sin->sin_family, &sin->sin_addr, buf4, sizeof(buf4))) {
			str4 = buf4;
		}
	}
	if (g_pconf->local_addr_diamid[1].ss_family == AF_INET6) {
		sSA6 * sin6 = (sSA6 *)&g_pconf->local_addr_diamid[1];
		if (inet_ntop(sin6->sin6_family, &sin6->sin6_addr, buf6, sizeof(buf6))) {
			str6 = buf6;
		}
	}
	
	TRACE_DEBUG( INFO, "--- Parsed configuration : ---");
	TRACE_DEBUG( INFO, "  Diameter-Id : %s ( %s / %s )", g_pconf->diameter_identity, str4, str6);
	TRACE_DEBUG( INFO, "  Local port  : %hd",g_pconf->local_port);
	TRACE_DEBUG( INFO, "  SCTP streams: %hd",g_pconf->sctp_streams);
	TRACE_DEBUG( INFO, "  Tc timer    : %d", g_pconf->tctimer);
	TRACE_DEBUG( INFO, "  Tw timer    : %d", g_pconf->twtimer);
	TRACE_DEBUG( INFO, "  Relaying    : %s", g_pconf->disable_relay ? "Disabled" : "Enabled");
	TRACE_DEBUG( INFO, "  TCP         : %s", g_pconf->disable_tcp   ? "Disabled" : "Enabled");
	TRACE_DEBUG( INFO, "  IP          : %s", g_pconf->disable_inet4 ? "Disabled" : "Enabled");
	TRACE_DEBUG( INFO, "  IPv6        : %s", g_pconf->disable_inet6 ? "Disabled" : "Enabled");
	   conf_dump_addr( "Primary",   g_pconf->local_addr_pri_sa );
	   conf_dump_addr( "Secondary", g_pconf->local_addr_sec_sa );
	TRACE_DEBUG( INFO, "------------------------------");
}

/* Initialize the configuration module */
int conf_init ( void )
{
	/* Set the global pointer */
	g_conf = &waaad_config;
	
	/* Now set the default values */
	return conf_setdefaults(g_conf);
}
	

/* Parse the configuration file (using the yacc parser) */
int conf_parse ( void )
{
	extern FILE * yyin;
	int ret;
	
	if (!g_conf) {
		TRACE_DEBUG (INFO, "Configuration is not initialized...");
		return EINVAL;
	}

	TRACE_DEBUG (FULL, "Parsing configuration file: %s...", g_conf->filepath);
	
	yyin = fopen(g_conf->filepath, "r");
	if (yyin == NULL) {
		ret = errno;
		log_error("Unable to open configuration file %s for reading: %s\n", g_conf->filepath, strerror(ret));
		TRACE_DEBUG (INFO, "Error occurred, message logged -- configuration file.");
		return ret;
	}
	
	/* call yacc parser */
	ret = yyparse(g_conf);

	fclose(yyin);

	if (ret != 0) {
		TRACE_DEBUG (INFO, "Unable to parse the configuration file.");
		return EINVAL;
	}
	
	/* resolve the diameter-id, store result in local_addr_diamid */
	{
		struct addrinfo *ai = NULL, *p, hints;
		int gotip4 = 0;
		int gotip6 = 0;
		
		memset(&hints, 0, sizeof(hints));
		hints.ai_flags = AI_ADDRCONFIG;

		ret = getaddrinfo(g_pconf->diameter_identity, NULL, &hints, &ai);
		if (ret != 0) {
			log_error("Error resolving Diameter-Id '%s': %s.\n", g_pconf->diameter_identity, gai_strerror(ret));
			return EINVAL;
		}
		for (p = ai; p != NULL; p = p->ai_next) {
			switch (p->ai_family) {
				case AF_INET:
					if (gotip4)
						break;
					memcpy(&g_pconf->local_addr_diamid[0], p->ai_addr, sizeof(sSA4));
					gotip4++;
					break;
					
				case AF_INET6:
					if (gotip6)
						break;
					memcpy(&g_pconf->local_addr_diamid[1], p->ai_addr, sizeof(sSA6));
					gotip6++;
					break;
			}
			if (gotip4 && gotip6)
				break;
		}

		freeaddrinfo(ai);
	}
	
	/* Dump the config and exit */
	conf_dump();

	return 0;
}	

/* Terminate the configuration module. */
int conf_fini ( void )
{
	if (g_pconf->diameter_identity)
		free(g_pconf->diameter_identity);
	
	if (g_pconf->local_addr_pri_sa)
		free(g_pconf->local_addr_pri_sa);
	
	if (g_pconf->local_addr_sec_sa)
		free(g_pconf->local_addr_sec_sa);
	
	return 0;
}
